{-# LANGUAGE OverloadedStrings #-}

module PUBMED where

import Prelude hiding (takeWhile)
import Data.Text (Text)
import PUBMED.Client
import PUBMED.Parser
import Network.HTTP.Client (newManager)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import Servant.Client (runClientM, mkClientEnv, BaseUrl(..), Scheme(..))
import Text.XML (parseLBS_, def)
import Text.XML.Cursor (fromDocument, Cursor)
import Data.Conduit (ConduitT)
import Data.ByteString.Lazy (ByteString)
import Data.ByteString.Char8 (pack)
import Control.Monad.Catch (MonadThrow)

import Control.Applicative
import Data.Attoparsec.ByteString
import Data.Attoparsec.ByteString.Char8 (anyChar)

import qualified Data.ByteString.Lazy as LBS
import qualified Data.ByteString      as DB
import qualified Data.Text as T

runParser :: Show res => (Cursor -> res) -> LBS.ByteString -> res
runParser parser = parser . fromDocument . parseLBS_ def

-- | TODO this parser need at least one subs at the end
-- (use endOfInput)
removeSub :: Parser ByteString
removeSub = do
  dt <- many textWithBalise
  eo <- manyTill anyChar endOfInput
  pure $ LBS.fromStrict $ pack $ concat dt <> eo
  where
    textWithBalise =
      manyTill anyChar (sub <|> asub)
    sub  = string "<sub>"
      <|> string "<sup>"
      <|> string "<i>"
    asub = string "</sub>"
      <|> string "</sup>"
      <|> string "</i>"

type Query = Text
type Limit = Integer

crawler :: Text -> Maybe Limit -> IO (Either Text [PubMed])
crawler = runSimpleFindPubmedAbstractRequest

runSimpleFindPubmedAbstractRequest :: Text -> Maybe Limit -> IO (Either Text [PubMed])
runSimpleFindPubmedAbstractRequest query limit =  do
  manager' <- newManager tlsManagerSettings
  res <- runClientM
    (search (Just query) limit)
    (mkClientEnv manager' $ BaseUrl Https "eutils.ncbi.nlm.nih.gov" 443 "entrez/eutils")
  case res of
    (Left err) -> pure (Left $ T.pack $ show err)
    (Right (BsXml docs)) -> do
      let docIds = runParser parseDocId docs
      res' <- runClientM
        (fetch (Just "pubmed") (Just "abstract") docIds)
        (mkClientEnv manager' $ BaseUrl Https "eutils.ncbi.nlm.nih.gov" 443 "entrez/eutils")
      case res' of
        (Left err') -> pure (Left $ T.pack $ show err')
        (Right (BsXml abstracts)) -> do
          -- TODO remove "</sub>" maybe there is a cleaner way with isEndOfInput
          case (parseOnly removeSub $ LBS.toStrict abstracts <> "</sub>") of
            (Left err'') -> pure (Left $ T.pack err'')
            (Right v) -> Right <$> pubMedParser v


